import os

from metaflow._vendor import click

from metaflow.extension_support.cmd import process_cmds, resolve_cmds
from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, CONTACT_INFO
from metaflow.metaflow_version import get_version

from .util import echo_always
import metaflow.tracing as tracing


@click.group()
@tracing.cli("cli/main")
def main():
    pass


@main.command(help="Show all available commands.")
@click.pass_context
def help(ctx):
    print(ctx.parent.get_help())


@main.command(help="Show flows accessible from the current working tree.")
def status():
    from metaflow.client import get_metadata

    res = get_metadata()
    if res:
        res = res.split("@")
    else:
        raise click.ClickException("Unknown status: cannot find a Metadata provider")
    if res[0] == "service":
        echo("Using Metadata provider at: ", nl=False)
        echo('"%s"\n' % res[1], fg="cyan")
        echo("To list available flows, type:\n")
        echo("1. python")
        echo("2. from metaflow import Metaflow")
        echo("3. list(Metaflow())")
        return

    from metaflow.client import namespace, metadata, Metaflow

    # Get the local data store path
    path = LocalStorage.get_datastore_root_from_config(echo, create_on_absent=False)
    # Throw an exception
    if path is None:
        raise click.ClickException(
            "Could not find "
            + click.style('"%s"' % DATASTORE_LOCAL_DIR, fg="red")
            + " in the current working tree."
        )

    stripped_path = os.path.dirname(path)
    namespace(None)
    metadata("local@%s" % stripped_path)
    echo("Working tree found at: ", nl=False)
    echo('"%s"\n' % stripped_path, fg="cyan")
    echo("Available flows:", fg="cyan", bold=True)
    for flow in Metaflow():
        echo("* %s" % flow, fg="cyan")


CMDS_DESC = [
    ("configure", ".configure_cmd.cli"),
    ("tutorials", ".tutorials_cmd.cli"),
    ("develop", ".develop.cli"),
    ("code", ".code.cli"),
]

process_cmds(globals())


@click.command(
    cls=click.CommandCollection,
    sources=[main] + resolve_cmds(),
    invoke_without_command=True,
)
@click.pass_context
def start(ctx):
    global echo
    echo = echo_always

    import metaflow

    version = get_version()
    echo("Metaflow ", fg="magenta", bold=True, nl=False)

    if ctx.invoked_subcommand is None:
        echo("(%s): " % version, fg="magenta", bold=False, nl=False)
    else:
        echo("(%s)\n" % version, fg="magenta", bold=False)

    if ctx.invoked_subcommand is None:
        echo("More AI, less engineering\n", fg="magenta")

        lnk_sz = max(len(lnk) for lnk in CONTACT_INFO.values()) + 1
        for what, lnk in CONTACT_INFO.items():
            echo("%s%s" % (lnk, " " * (lnk_sz - len(lnk))), fg="cyan", nl=False)
            echo("- %s" % what)
        echo("")

        print(ctx.get_help())


if __name__ == "__main__":
    start()



from metaflow.exception import MetaflowException
from metaflow.plugins.secrets.utils import get_default_secrets_backend_type


class SecretSpec:
    def __init__(self, secrets_backend_type, secret_id, options={}, role=None):
        self._secrets_backend_type = secrets_backend_type
        self._secret_id = secret_id
        self._options = options
        self._role = role

    @property
    def secrets_backend_type(self):
        return self._secrets_backend_type

    @property
    def secret_id(self):
        return self._secret_id

    @property
    def options(self):
        return self._options

    @property
    def role(self):
        return self._role

    def to_json(self):
        """Mainly used for testing... not the same as the input dict in secret_spec_from_dict()!"""
        return {
            "secrets_backend_type": self.secrets_backend_type,
            "secret_id": self.secret_id,
            "options": self.options,
            "role": self.role,
        }

    def __str__(self):
        return "%s (%s)" % (self._secret_id, self._secrets_backend_type)

    @staticmethod
    def secret_spec_from_str(secret_spec_str, role):
        # "." may be used in secret_id one day (provider specific). HOWEVER, it provides the best UX for
        # non-conflicting cases (i.e. for secret ids that don't contain "."). This is true for all AWS
        # Secrets Manager secrets.
        #
        # So we skew heavily optimize for best upfront UX for the present (1/2023).
        #
        # If/when a certain secret backend supports "." secret names, we can figure out a solution at that time.
        # At a minimum, dictionary style secret spec may be used with no code changes (see secret_spec_from_dict()).
        # Other options could be:
        #  - accept and document that "." secret_ids don't work in Metaflow (across all possible providers)
        #  - add a Metaflow config variable that specifies the separator (default ".")
        #  - smarter spec parsing, that errors on secrets that look ambiguous. "aws-secrets-manager.XYZ" could mean:
        #    + secret_id "XYZ" in aws-secrets-manager backend, OR
        #    + secret_id "aws-secrets-manager.XYZ" default backend (if it is defined).
        #    + in this case, user can simply set "azure-key-vault.aws-secrets-manager.XYZ" instead!
        parts = secret_spec_str.split(".", maxsplit=1)
        if len(parts) == 1:
            secrets_backend_type = get_default_secrets_backend_type()
            secret_id = parts[0]
        else:
            secrets_backend_type = parts[0]
            secret_id = parts[1]
        return SecretSpec(
            secrets_backend_type, secret_id=secret_id, options={}, role=role
        )

    @staticmethod
    def secret_spec_from_dict(secret_spec_dict, role):
        if "type" not in secret_spec_dict:
            secrets_backend_type = get_default_secrets_backend_type()
        else:
            secrets_backend_type = secret_spec_dict["type"]
            if not isinstance(secrets_backend_type, str):
                raise MetaflowException(
                    "Bad @secrets specification - 'type' must be a string - found %s"
                    % type(secrets_backend_type)
                )
        secret_id = secret_spec_dict.get("id")
        if not isinstance(secret_id, str):
            raise MetaflowException(
                "Bad @secrets specification - 'id' must be a string - found %s"
                % type(secret_id)
            )
        options = secret_spec_dict.get("options", {})
        if not isinstance(options, dict):
            raise MetaflowException(
                "Bad @secrets specification - 'option' must be a dict - found %s"
                % type(options)
            )
        role_for_source = secret_spec_dict.get("role", None)
        if role_for_source is not None:
            if not isinstance(role_for_source, str):
                raise MetaflowException(
                    "Bad @secrets specification - 'role' must be a str - found %s"
                    % type(role_for_source)
                )
            role = role_for_source
        return SecretSpec(
            secrets_backend_type, secret_id=secret_id, options=options, role=role
        )



import inspect
import json
import re
import types

from typing import Any, Callable, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

from metaflow.debug import debug
from metaflow.exception import MetaflowException
from metaflow.user_configs.config_parameters import (
    resolve_delayed_evaluator,
    unpack_delayed_evaluator,
)

from .common import ClassPath_Trie

if TYPE_CHECKING:
    import metaflow.datastore.inputs
    import metaflow.decorators
    import metaflow.flowspec
    import metaflow.user_decorators.mutable_step

USER_SKIP_STEP = {}


class UserStepDecoratorMeta(type):
    _all_registered_decorators = ClassPath_Trie()
    _do_not_register = set()
    _import_modules = set()

    def __new__(mcs, name, bases, namespace, **_kwargs):
        cls = super().__new__(mcs, name, bases, namespace)
        cls.decorator_name = getattr(
            cls, "_decorator_name", f"{cls.__module__}.{cls.__name__}"
        )
        effective_module = getattr(cls, "_original_module", cls.__module__)
        if not effective_module.startswith(
            "metaflow."
        ) and not effective_module.startswith("metaflow_extensions."):
            mcs._import_modules.add(effective_module)

        if (
            name in ("FlowMutator", "UserStepDecorator")
            or cls.decorator_name in mcs._do_not_register
        ):
            return cls

        # We inject a __init_subclass__ method so we can figure out if there
        # are subclasses. We want to register as decorators only the ones that do
        # not have a subclass. The logic is that everything is registered and if
        # a subclass shows up, we will unregister the parent class leaving only those
        # classes that do not have any subclasses registered.
        @classmethod
        def do_unregister(cls_, **_kwargs):
            for base in cls_.__bases__:
                if isinstance(base, UserStepDecoratorMeta):
                    # If the base is a UserStepDecoratorMeta, we unregister it
                    # so that we don't have any decorators that are not the
                    # most derived one.
                    mcs._all_registered_decorators.remove(base.decorator_name)
                    # Also make sure we don't register again
                    mcs._do_not_register.add(base.decorator_name)

        cls.__init_subclass__ = do_unregister
        mcs._all_registered_decorators.insert(cls.decorator_name, cls)
        return cls

    def __str__(cls):
        return "%s(%s)" % (
            cls.__name__ if cls.__name__ != "WrapClass" else "UserStepDecorator",
            getattr(cls, "decorator_name", None),
        )

    @classmethod
    def all_decorators(mcs) -> Dict[str, "UserStepDecoratorMeta"]:
        """
        Get all registered decorators using the minimally unique classpath name

        Returns
        -------
        Dict[str, UserStepDecoratorBase]
            A dictionary mapping decorator names to their classes.
        """
        mcs._check_init()
        return mcs._all_registered_decorators.get_unique_prefixes()

    @classmethod
    def get_decorator_by_name(
        mcs, decorator_name: str
    ) -> Optional[Union["UserStepDecoratorBase", "metaflow.decorators.Decorator"]]:
        """
        Get a decorator by its name.

        Parameters
        ----------
        decorator_name: str
            The name of the decorator to retrieve.

        Returns
        -------
        Optional[UserStepDecoratorBase]
            The decorator class if found, None otherwise.
        """
        mcs._check_init()
        return mcs._all_registered_decorators.unique_prefix_value(decorator_name)

    @classmethod
    def get_decorator_name(mcs, decorator_type: type) -> Optional[str]:
        """
        Get the minimally unique classpath name for a decorator type.

        Parameters
        ----------
        decorator_type: type
            The type of the decorator to retrieve the name for.

        Returns
        -------
        Optional[str]
            The minimally unique classpath name if found, None otherwise.
        """
        mcs._check_init()
        return mcs._all_registered_decorators.unique_prefix_for_type(decorator_type)

    @classmethod
    def _check_init(mcs):
        # Delay importing STEP_DECORATORS until we actually need it
        if not mcs._all_registered_decorators.inited:
            from metaflow.plugins import STEP_DECORATORS

            mcs._all_registered_decorators.init([(t.name, t) for t in STEP_DECORATORS])


class UserStepDecoratorBase(metaclass=UserStepDecoratorMeta):
    _step_field = None
    _allowed_args = False
    _allowed_kwargs = False

    def __init__(self, *args, **kwargs):
        arg = None
        self._args = args
        self._kwargs = {}
        # If nothing is set, the user statically defined the decorator
        self._special_kwargs = {"_statically_defined": True, "_inserted_by": None}
        for k, v in kwargs.items():
            if k in ("_statically_defined", "_inserted_by"):
                # These are special arguments that we do not want to pass to the step
                # decorator
                self._special_kwargs[k] = v
            else:
                self._kwargs[k] = v

        if self._args:
            if isinstance(self._args[0], UserStepDecoratorBase):
                arg = self._args[0]._my_step
            else:
                arg = self._args[0]

        if arg and callable(arg) and hasattr(arg, "is_step"):
            # This means the decorator is bare like @MyDecorator
            # and the first argument is the step
            self._set_my_step(arg)
            self._args = args[1:]  # The rest of the args are the decorator args

        if self._args and not self._allowed_args:
            raise MetaflowException("%s does not allow arguments" % str(self))
        if self._kwargs:
            if not self._allowed_kwargs:
                raise MetaflowException("%s does not allow keyword arguments" % self)
            elif isinstance(self._allowed_kwargs, list) and any(
                a not in self._allowed_kwargs for a in self._kwargs
            ):
                raise MetaflowException(
                    "%s only allows the following keyword arguments: %s"
                    % (self, str(self._allowed_args))
                )

    def __get__(self, instance, owner):
        # Required so that we "present" as a step when the step decorator is
        # of the form
        # @MyStepDecorator
        # @step
        # def my_step(self):
        #     pass
        #
        # This is *not* called for something like:
        # @MyStepDecorator()
        # @step
        # def my_step(self):
        #     pass
        # because in that case, we will have called __call__ below and that already
        # returns a function and that __get__ function will be called.

        return self().__get__(instance, owner)

    def __call__(
        self,
        step: Optional[
            Union[
                Callable[["metaflow.decorators.FlowSpecDerived"], None],
                Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
            ]
        ] = None,
        **kwargs,
    ) -> Union[
        Callable[["metaflow.decorators.FlowSpecDerived"], None],
        Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
    ]:
        # The only kwargs here are just special kwargs (not user facing since those
        # are passed in the constructor)
        self._special_kwargs.update(kwargs)
        if step:
            if isinstance(step, UserStepDecoratorBase):
                step = step._my_step

            return self._set_my_step(step)
        elif not self._my_step:
            # This means that somehow the initialization did not happen properly
            # so this may have been applied to a non step
            raise MetaflowException("%s can only be applied to a step function" % self)
        return self._my_step

    def add_or_raise(
        self,
        step: Union[
            Callable[["metaflow.decorators.FlowSpecDerived"], None],
            Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
        ],
        statically_defined: bool,
        duplicates: int,
        inserted_by: Optional[str] = None,
    ):
        from metaflow.user_decorators.mutable_step import MutableStep

        existing_deco = [
            d
            for d in getattr(step, self._step_field)
            if d.decorator_name == self.decorator_name
        ]

        if not existing_deco:
            self(step, _statically_defined=statically_defined, _inserted_by=inserted_by)
        elif duplicates == MutableStep.IGNORE:
            # If we are ignoring duplicates, we just return
            debug.userconf_exec(
                "Ignoring duplicate decorator %s on step %s from %s"
                % (self, step.__name__, inserted_by)
            )
            return
        elif duplicates == MutableStep.OVERRIDE:
            # If we are overriding, we remove the existing decorator and add this one
            debug.userconf_exec(
                "Overriding decorator %s on step %s from %s"
                % (self, step.__name__, inserted_by)
            )
            setattr(
                step,
                self._step_field,
                [
                    d
                    for d in getattr(step, self._step_field)
                    if d.decorator_name != self.decorator_name
                ],
            )
            self(step, _statically_defined=statically_defined, _inserted_by=inserted_by)
        elif duplicates == MutableStep.ERROR:
            if statically_defined:
                # Prevent circular dep
                from metaflow.decorators import DuplicateStepDecoratorException

                raise DuplicateStepDecoratorException(self.__class__, step)

    def _set_my_step(
        self,
        step: Union[
            Callable[["metaflow.decorators.FlowSpecDerived"], None],
            Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
        ],
    ) -> Union[
        Callable[["metaflow.decorators.FlowSpecDerived"], None],
        Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
    ]:
        self._my_step = step
        if self._step_field is None:
            raise RuntimeError(
                "UserStepDecorator is not properly overloaded; missing _step_field. "
                "This is a Metaflow bug, please contact support."
            )
        # When we set the step, we can now determine if we are statically defined or
        # not. We can't do it much earlier because the decorator itself may be defined
        # (ie: @user_step_decorator is statically defined) but it will only be a static
        # decorator when the user applies it to a step function.
        self.statically_defined = self._special_kwargs["_statically_defined"]
        self.inserted_by = self._special_kwargs["_inserted_by"]

        getattr(self._my_step, self._step_field).append(self)
        return self._my_step

    def __str__(self):
        return str(self.__class__)

    @classmethod
    def extract_args_kwargs_from_decorator_spec(
        cls, deco_spec: str
    ) -> Tuple[List[Any], Dict[str, Any]]:
        if len(deco_spec) == 0:
            return [], {}
        args = []
        kwargs = {}
        for a in re.split(r""",(?=[\s\w]+=)""", deco_spec):
            name, val = a.split("=", 1)
            try:
                val_parsed = json.loads(val.strip().replace('\\"', '"'))
            except json.JSONDecodeError:
                # In this case, we try to convert to either an int or a float or
                # leave as is. Prefer ints if possible.
                try:
                    val_parsed = int(val.strip())
                except ValueError:
                    try:
                        val_parsed = float(val.strip())
                    except ValueError:
                        val_parsed = val.strip()
            try:
                pos = int(name)
            except ValueError:
                kwargs[name.strip()] = val_parsed
            else:
                # Extend args list if needed to accommodate position
                while len(args) <= pos:
                    args.append(None)
                args[pos] = val_parsed
        debug.userconf_exec(
            "Parsed decorator spec for %s: %s"
            % (cls.decorator_name, str((args, kwargs)))
        )
        return args, kwargs

    @classmethod
    def parse_decorator_spec(cls, deco_spec: str) -> Optional["UserStepDecoratorBase"]:
        if len(deco_spec) == 0:
            return cls()
        args, kwargs = cls.extract_args_kwargs_from_decorator_spec(deco_spec)
        return cls(*args, **kwargs)

    def make_decorator_spec(self):
        self.external_init()
        attrs = {}
        if self._args:
            attrs.update({i: v for i, v in enumerate(self._args) if v is not None})
        if self._kwargs:
            attrs.update({k: v for k, v in self._kwargs.items() if v is not None})
        if attrs:
            attr_list = []
            # We dump simple types directly as string to get around the nightmare quote
            # escaping but for more complex types (typically dictionaries or lists),
            # we dump using JSON.
            for k, v in attrs.items():
                if isinstance(v, (int, float, str)):
                    attr_list.append("%s=%s" % (k, str(v)))
                else:
                    attr_list.append("%s=%s" % (k, json.dumps(v).replace('"', '\\"')))

            attrstr = ",".join(attr_list)
            return "%s:%s" % (self.decorator_name, attrstr)
        else:
            return self.decorator_name

    def get_args_kwargs(self) -> Tuple[List[Any], Dict[str, Any]]:
        """
        Get the arguments and keyword arguments of the decorator.

        Returns
        -------
        Tuple[List[Any], Dict[str, Any]]
            A tuple containing a list of arguments and a dictionary of keyword arguments.
        """
        return list(self._args), dict(self._kwargs)

    def init(self, *args, **kwargs):
        pass

    def external_init(self):
        # You can use config values in the arguments to a UserStepDecoratorBase
        # so we resolve those as well
        self._args = [resolve_delayed_evaluator(arg) for arg in self._args]
        self._kwargs, _ = unpack_delayed_evaluator(self._kwargs)
        self._kwargs = {
            k: resolve_delayed_evaluator(v) for k, v in self._kwargs.items()
        }
        if self._args or self._kwargs:
            if "init" not in self.__class__.__dict__:
                raise MetaflowException(
                    "%s is used with arguments but does not implement init" % self
                )
        if "init" in self.__class__.__dict__:
            self.init(*self._args, **self._kwargs)


class UserStepDecorator(UserStepDecoratorBase):
    _step_field = "wrappers"
    _allowed_args = False
    _allowed_kwargs = True

    def init(self, *args, **kwargs):
        """
        Implement this method if your UserStepDecorator takes arguments. It replaces the
        __init__ method in traditional Python classes.


        As an example:
        ```
        class MyDecorator(UserStepDecorator):
            def init(self, *args, **kwargs):
                self.arg1 = kwargs.get("arg1", None)
                self.arg2 = kwargs.get("arg2", None)
                # Do something with the arguments
        ```

        can then be used as
        ```
        @MyDecorator(arg1=42, arg2=conf_expr("config.my_arg2"))
        @step
        def start(self):
            pass
        ```
        """
        super().init()

    def pre_step(
        self,
        step_name: str,
        flow: "metaflow.flowspec.FlowSpec",
        inputs: Optional["metaflow.datastore.inputs.Inputs"] = None,
    ) -> Optional[Callable[["metaflow.flowspec.FlowSpec", Optional[Any]], Any]]:
        """
        Implement this method to perform any action prior to the execution of a step.

        It should return either None to execute anything wrapped by this step decorator
        as usual or a callable that will be called instead.

        Parameters
        ----------
        step_name: str
            The name of the step being decorated.
        flow: FlowSpec
            The flow object to which the step belongs.
        inputs: Optional[List[FlowSpec]]
            The inputs to the step being decorated. This is only provided for join steps
            and is None for all other steps.

        Returns
        -------
        Optional[Callable[FlowSpec, Optional[Any]]]
            An optional function to use instead of the wrapped step. Note that the function
            returned should match the signature of the step being wrapped (join steps
            take an additional "inputs" argument).
        """
        return None

    def post_step(
        self,
        step_name: str,
        flow: "metaflow.flowspec.FlowSpec",
        exception: Optional[Exception] = None,
    ) -> Optional[
        Union[Optional[Exception], Tuple[Optional[Exception], Optional[Dict[str, Any]]]]
    ]:
        """
        Implement this method to perform any action after the execution of a step.

        If the step (or any code being wrapped by this decorator) raises an exception,
        it will be passed here and can either be caught (in which case the step will
        be considered as successful) or re-raised (in which case the entire step
        will be considered a failure unless another decorator catches the execption).

        Note that this method executes *before* artifacts are stored in the datastore
        so it is able to modify, add or remove artifacts from `flow`.

        Parameters
        ----------
        step_name: str
            The name of the step being decorated.
        flow: FlowSpec
            The flow object to which the step belongs.
        exception: Optional[Exception]
            The exception raised during the step execution, if any.

        Returns
        -------
        Optional[Union[Optional[Exception], Tuple[Optional[Exception], Optional[Dict[str, Any]]]]]
            An exception (if None, the step is considered successful)
            OR
            A tuple containing:
              - An exception to be raised (if None, the step is considered successful).
              - A dictionary with values to pass to `self.next()`. If an empty dictionary
                is returned, the default arguments to `self.next()` for this step will be
                used. Return None if you do not want to call `self.next()` at all
                (this is typically the case as the step will call it itself).
        Note that returning None will gobble the exception.
        """
        if exception:
            return exception, None
        return None, None

    @property
    def skip_step(self) -> Union[bool, Dict[str, Any]]:
        """
        Returns whether or not the step (or rather anything wrapped by this decorator)
        should be skipped

        Returns
        -------
        Union[bool, Dict[str, Any]]
            False if the step should not be skipped. True if it should be skipped and
            a dictionary if it should be skipped and the values passed in used as
            the arguments to the self.next call.
        """
        return getattr(self, "_skip_step", False)

    @skip_step.setter
    def skip_step(self, value: Union[bool, Dict[str, Any]]):
        """
        Set the skip_step property. You can set it to:
          - True to skip the step
          - False to not skip the step (default)
          - A dictionary with the keys valid in the `self.next` call.

        Parameters
        ----------
        value: Union[bool, Dict[str, Any]]
            True/False or a dictionary with the keys valid in the `self.next` call.
        """
        self._skip_step = value


def user_step_decorator(*args, **kwargs):
    """
    Use this decorator to transform a generator function into a user step decorator.

    As an example:

    ```
    @user_step_decorator
    def timing(step_name, flow, inputs):
        start_time = time.time()
        yield
        end_time = time.time()
        flow.artifact_total_time = end_time - start_time
        print(f"Step {step_name} took {flow.artifact_total_time} seconds")
    ```
    which can then be used as:

    ```
    @timing
    @step
    def start(self):
        print("Hello, world!")
    ```

    Your generator should:
      - yield at most once -- if you do not yield, the step will not execute.
      - yield:
          - None
          - a callable that will replace whatever is being wrapped (it
            should have the same parameters as the wrapped function, namely, it should
            be a
            Callable[[FlowSpec, Inputs], Optional[Union[Dict[str, Any]]]]).
            Note that the return type is a bit different -- you can return:
              - None: no special behavior;
              - A dictionary containing parameters for `self.next()`.
          - a dictionary to skip the step. An empty dictionary is equivalent
            to just skipping the step. A full dictionary will pass the arguments
            to the `self.next()` call -- this allows you to modify the behavior
            of `self.next` (for example, changing the `foreach` values. We provide
            USER_SKIP_STEP as a special value that is equivalent to {}.


    You are able to catch exceptions thrown by the yield statement (ie: coming from the
    wrapped code). Catching and not re-raising the exception will make the step
    successful.

    Note that you are able to modify the step's artifact after the yield.

    For more complex use cases, you can use the `UserStepDecorator` class directly which
    allows more control.
    """
    if args:
        # If we have args, we either had @user_step_decorator with no argument or we had
        # @user_step_decorator(arg="foo") and transformed it into
        # @user_step_decorator(step, arg="foo")
        obj = args[0]
        name = f"{obj.__module__}.{obj.__name__}"

        if not isinstance(obj, types.FunctionType) or not inspect.isgeneratorfunction(
            obj
        ):
            raise MetaflowException(
                "@user_step_decorator can only decorate generator functions."
            )
        sig = inspect.signature(obj)
        arg_count = len(sig.parameters)
        if kwargs:
            if arg_count != 4:
                raise MetaflowException(
                    "@user_step_decorator(<kwargs>) can only decorate generator "
                    "functions with 4 arguments (step_name, flow, inputs, attributes)"
                )
        elif arg_count not in (3, 4):
            raise MetaflowException(
                "@user_step_decorator can only decorator generator functions with 3 or "
                "4 arguments (step_name, flow, inputs [, attributes])."
            )

        class WrapClass(UserStepDecorator):
            _allowed_args = False
            _allowed_kwargs = True
            _step_field = "wrappers"
            _decorator_name = name
            _original_module = obj.__module__

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._generator = obj

            def init(self, *args, **kwargs):
                if args:
                    raise MetaflowException(
                        "%s does not allow arguments, only keyword arguments"
                        % str(self)
                    )
                self._kwargs = kwargs

            def pre_step(self, step_name, flow, inputs):
                if arg_count == 4:
                    self._generator = self._generator(
                        step_name, flow, inputs, self._kwargs
                    )
                else:
                    self._generator = self._generator(step_name, flow, inputs)
                v = self._generator.send(None)
                if isinstance(v, dict):
                    # We are modifying the behavior of self.next
                    if v:
                        self.skip_step = v
                    else:
                        # Emtpy dict is just skip the step
                        self.skip_step = True
                    return None
                return v

            def post_step(self, step_name, flow, exception=None):
                to_return = None, None
                try:
                    if exception:
                        self._generator.throw(exception)
                    else:
                        self._generator.send(None)
                except StopIteration as e:
                    to_return = e.value
                except Exception as e:
                    return e
                else:
                    return (
                        None,
                        None,
                        MetaflowException(" %s should only yield once" % self),
                    )
                return to_return

        return WrapClass
    else:
        # Capture arguments passed to user_step_decorator
        def wrap(f):
            return user_step_decorator(f, **kwargs)

        return wrap


class StepMutator(UserStepDecoratorBase):
    """
    Derive from this class to implement a step mutator.

    A step mutator allows you to introspect a step and add decorators to it. You can
    use values available through configurations to determine how to mutate the step.

    There are two main methods provided:
      - pre_mutate: called as early as possible right after configuration values are read.
      - mutate: called right after all the command line is parsed but before any
        Metaflow decorators are applied.
    """

    _step_field = "config_decorators"
    _allowed_args = True
    _allowed_kwargs = True

    def init(self, *args, **kwargs):
        """
        Implement this method if you wish for your StepMutator to take in arguments.

        Your step-mutator can then look like:

        @MyMutator(arg1, arg2)
        @step
        def my_step(self):
            pass

        It is an error to use your mutator with arguments but not implement this method.
        """
        super().init()

    def pre_mutate(
        self, mutable_step: "metaflow.user_decorators.mutable_step.MutableStep"
    ) -> None:
        """
        Method called right after all configuration values are read.

        Parameters
        ----------
        mutable_step : metaflow.user_decorators.mutable_step.MutableStep
            A representation of this step
        """
        return None

    def mutate(
        self, mutable_step: "metaflow.user_decorators.mutable_step.MutableStep"
    ) -> None:
        """
        Method called right before the first Metaflow decorator is applied. This
        means that the command line, including all `--with` options has been parsed.

        Parameters
        ----------
        mutable_step : metaflow.user_decorators.mutable_step.MutableStep
            A representation of this step
        """
        return None



import hashlib
import json
import sys
import platform
from collections import defaultdict
from datetime import datetime, timedelta


TASK_ID_XCOM_KEY = "metaflow_task_id"
FOREACH_CARDINALITY_XCOM_KEY = "metaflow_foreach_cardinality"
FOREACH_XCOM_KEY = "metaflow_foreach_indexes"
RUN_HASH_ID_LEN = 12
TASK_ID_HASH_LEN = 8
RUN_ID_PREFIX = "airflow"
AIRFLOW_FOREACH_SUPPORT_VERSION = "2.3.0"
AIRFLOW_MIN_SUPPORT_VERSION = "2.2.0"
KUBERNETES_PROVIDER_FOREACH_VERSION = "4.2.0"


class KubernetesProviderNotFound(Exception):
    headline = "Kubernetes provider not found"


class ForeachIncompatibleException(Exception):
    headline = "Airflow version is incompatible to support Metaflow `foreach`s."


class IncompatibleVersionException(Exception):
    headline = "Metaflow is incompatible with current version of Airflow."

    def __init__(self, version_number) -> None:
        msg = (
            "Airflow version %s is incompatible with Metaflow. Metaflow requires Airflow a minimum version %s"
            % (version_number, AIRFLOW_MIN_SUPPORT_VERSION)
        )
        super().__init__(msg)


class IncompatibleKubernetesProviderVersionException(Exception):
    headline = (
        "Kubernetes Provider version is incompatible with Metaflow `foreach`s. "
        "Install the provider via "
        "`%s -m pip install apache-airflow-providers-cncf-kubernetes==%s`"
    ) % (sys.executable, KUBERNETES_PROVIDER_FOREACH_VERSION)


class AirflowSensorNotFound(Exception):
    headline = "Sensor package not found"


def create_absolute_version_number(version):
    abs_version = None
    # For all digits
    if all(v.isdigit() for v in version.split(".")):
        abs_version = sum(
            [
                (10 ** (3 - idx)) * i
                for idx, i in enumerate([int(v) for v in version.split(".")])
            ]
        )
    # For first two digits
    elif all(v.isdigit() for v in version.split(".")[:2]):
        abs_version = sum(
            [
                (10 ** (3 - idx)) * i
                for idx, i in enumerate([int(v) for v in version.split(".")[:2]])
            ]
        )
    return abs_version


def _validate_dynamic_mapping_compatibility():
    from airflow.version import version

    af_ver = create_absolute_version_number(version)
    if af_ver is None or af_ver < create_absolute_version_number(
        AIRFLOW_FOREACH_SUPPORT_VERSION
    ):
        ForeachIncompatibleException(
            "Please install airflow version %s to use Airflow's Dynamic task mapping functionality."
            % AIRFLOW_FOREACH_SUPPORT_VERSION
        )


def get_kubernetes_provider_version():
    try:
        from airflow.providers.cncf.kubernetes.get_provider_info import (
            get_provider_info,
        )
    except ImportError as e:
        raise KubernetesProviderNotFound(
            "This DAG utilizes `KubernetesPodOperator`. "
            "Install the Airflow Kubernetes provider using "
            "`%s -m pip install apache-airflow-providers-cncf-kubernetes`"
            % sys.executable
        )
    return get_provider_info()["versions"][0]


def _validate_minimum_airflow_version():
    from airflow.version import version

    af_ver = create_absolute_version_number(version)
    if af_ver is None or af_ver < create_absolute_version_number(
        AIRFLOW_MIN_SUPPORT_VERSION
    ):
        raise IncompatibleVersionException(version)


def _check_foreach_compatible_kubernetes_provider():
    provider_version = get_kubernetes_provider_version()
    ver = create_absolute_version_number(provider_version)
    if ver is None or ver < create_absolute_version_number(
        KUBERNETES_PROVIDER_FOREACH_VERSION
    ):
        raise IncompatibleKubernetesProviderVersionException()


def datetimeparse(isotimestamp):
    ver = int(platform.python_version_tuple()[0]) * 10 + int(
        platform.python_version_tuple()[1]
    )
    if ver >= 37:
        return datetime.fromisoformat(isotimestamp)
    else:
        return datetime.strptime(isotimestamp, "%Y-%m-%dT%H:%M:%S.%f")


def get_xcom_arg_class():
    try:
        from airflow import XComArg
    except ImportError:
        return None
    return XComArg


class AIRFLOW_MACROS:
    # run_id_creator is added via the `user_defined_filters`
    RUN_ID = "%s-{{ [run_id, dag_run.dag_id] | run_id_creator }}" % RUN_ID_PREFIX
    PARAMETERS = "{{ params | json_dump }}"

    STEPNAME = "{{ ti.task_id }}"

    # AIRFLOW_MACROS.TASK_ID will work for linear/branched workflows.
    # ti.task_id is the stepname in metaflow code.
    # AIRFLOW_MACROS.TASK_ID uses a jinja filter called `task_id_creator` which helps
    # concatenate the string using a `/`. Since run-id will keep changing and stepname will be
    # the same task id will change. Since airflow doesn't encourage dynamic rewriting of dags
    # we can rename steps in a foreach with indexes (eg. `stepname-$index`) to create those steps.
    # Hence : `foreach`s will require some special form of plumbing.
    # https://stackoverflow.com/questions/62962386/can-an-airflow-task-dynamically-generate-a-dag-at-runtime
    TASK_ID = (
        "%s-{{ [run_id, ti.task_id, dag_run.dag_id] | task_id_creator  }}"
        % RUN_ID_PREFIX
    )

    FOREACH_TASK_ID = (
        "%s-{{ [run_id, ti.task_id, dag_run.dag_id, ti.map_index] | task_id_creator  }}"
        % RUN_ID_PREFIX
    )

    # Airflow run_ids are of the form : "manual__2022-03-15T01:26:41.186781+00:00"
    # Such run-ids break the `metaflow.util.decompress_list`; this is why we hash the runid
    # We do `echo -n` because it emits line breaks, and we don't want to consider that, since we want same hash value
    # when retrieved in python.
    RUN_ID_SHELL = (
        "%s-$(echo -n {{ run_id }}-{{ dag_run.dag_id }} | md5sum | awk '{print $1}' | awk '{print substr ($0, 0, %s)}')"
        % (RUN_ID_PREFIX, str(RUN_HASH_ID_LEN))
    )

    ATTEMPT = "{{ task_instance.try_number - 1 }}"

    AIRFLOW_RUN_ID = "{{ run_id }}"

    AIRFLOW_JOB_ID = "{{ ti.job_id }}"

    FOREACH_SPLIT_INDEX = "{{ ti.map_index }}"

    @classmethod
    def create_task_id(cls, is_foreach):
        if is_foreach:
            return cls.FOREACH_TASK_ID
        else:
            return cls.TASK_ID

    @classmethod
    def pathspec(cls, flowname, is_foreach=False):
        return "%s/%s/%s/%s" % (
            flowname,
            cls.RUN_ID,
            cls.STEPNAME,
            cls.create_task_id(is_foreach),
        )


class SensorNames:
    EXTERNAL_TASK_SENSOR = "ExternalTaskSensor"
    S3_SENSOR = "S3KeySensor"

    @classmethod
    def get_supported_sensors(cls):
        return list(cls.__dict__.values())


def run_id_creator(val):
    # join `[dag-id,run-id]` of airflow dag.
    return hashlib.md5("-".join([str(x) for x in val]).encode("utf-8")).hexdigest()[
        :RUN_HASH_ID_LEN
    ]


def task_id_creator(val):
    # join `[dag-id,run-id]` of airflow dag.
    return hashlib.md5("-".join([str(x) for x in val]).encode("utf-8")).hexdigest()[
        :TASK_ID_HASH_LEN
    ]


def id_creator(val, hash_len):
    # join `[dag-id,run-id]` of airflow dag.
    return hashlib.md5("-".join([str(x) for x in val]).encode("utf-8")).hexdigest()[
        :hash_len
    ]


def json_dump(val):
    return json.dumps(val)


class AirflowDAGArgs(object):

    # `_arg_types` is a dictionary which represents the types of the arguments of an Airflow `DAG`.
    # `_arg_types` is used when parsing types back from the configuration json.
    # It doesn't cover all the arguments but covers many of the important one which can come from the cli.
    _arg_types = {
        "dag_id": str,
        "description": str,
        "schedule_interval": str,
        "start_date": datetime,
        "catchup": bool,
        "tags": list,
        "dagrun_timeout": timedelta,
        "default_args": {
            "owner": str,
            "depends_on_past": bool,
            "email": list,
            "email_on_failure": bool,
            "email_on_retry": bool,
            "retries": int,
            "retry_delay": timedelta,
            "queue": str,  # which queue to target when running this job. Not all executors implement queue management, the CeleryExecutor does support targeting specific queues.
            "pool": str,  # the slot pool this task should run in, slot pools are a way to limit concurrency for certain tasks
            "priority_weight": int,
            "wait_for_downstream": bool,
            "sla": timedelta,
            "execution_timeout": timedelta,
            "trigger_rule": str,
        },
    }

    # Reference for user_defined_filters : https://stackoverflow.com/a/70175317
    filters = dict(
        task_id_creator=lambda v: task_id_creator(v),
        json_dump=lambda val: json_dump(val),
        run_id_creator=lambda val: run_id_creator(val),
        join_list=lambda x: ",".join(list(x)),
    )

    def __init__(self, **kwargs):
        self._args = kwargs

    @property
    def arguments(self):
        return dict(**self._args, user_defined_filters=self.filters)

    def serialize(self):
        def parse_args(dd):
            data_dict = {}
            for k, v in dd.items():
                if isinstance(v, dict):
                    data_dict[k] = parse_args(v)
                elif isinstance(v, datetime):
                    data_dict[k] = v.isoformat()
                elif isinstance(v, timedelta):
                    data_dict[k] = dict(seconds=v.total_seconds())
                else:
                    data_dict[k] = v
            return data_dict

        return parse_args(self._args)

    @classmethod
    def deserialize(cls, data_dict):
        def parse_args(dd, type_check_dict):
            kwrgs = {}
            for k, v in dd.items():
                if k not in type_check_dict:
                    kwrgs[k] = v
                elif isinstance(v, dict) and isinstance(type_check_dict[k], dict):
                    kwrgs[k] = parse_args(v, type_check_dict[k])
                elif type_check_dict[k] == datetime:
                    kwrgs[k] = datetimeparse(v)
                elif type_check_dict[k] == timedelta:
                    kwrgs[k] = timedelta(**v)
                else:
                    kwrgs[k] = v
            return kwrgs

        return cls(**parse_args(data_dict, cls._arg_types))


def _kubernetes_pod_operator_args(operator_args):
    from kubernetes import client

    from airflow.kubernetes.secret import Secret

    # Set dynamic env variables like run-id, task-id etc from here.
    secrets = [
        Secret("env", secret, secret) for secret in operator_args.get("secrets", [])
    ]
    args = operator_args
    args.update(
        {
            "secrets": secrets,
            # Question for (savin):
            # Default timeout in airflow is 120. I can remove `startup_timeout_seconds` for now. how should we expose it to the user?
        }
    )
    # We need to explicitly add the `client.V1EnvVar` over here because
    # `pod_runtime_info_envs` doesn't accept arguments in dictionary form and strictly
    # Requires objects of type `client.V1EnvVar`
    additional_env_vars = [
        client.V1EnvVar(
            name=k,
            value_from=client.V1EnvVarSource(
                field_ref=client.V1ObjectFieldSelector(field_path=str(v))
            ),
        )
        for k, v in {
            "METAFLOW_KUBERNETES_POD_NAMESPACE": "metadata.namespace",
            "METAFLOW_KUBERNETES_POD_NAME": "metadata.name",
            "METAFLOW_KUBERNETES_POD_ID": "metadata.uid",
            "METAFLOW_KUBERNETES_SERVICE_ACCOUNT_NAME": "spec.serviceAccountName",
            "METAFLOW_KUBERNETES_NODE_IP": "status.hostIP",
        }.items()
    ]
    args["pod_runtime_info_envs"] = additional_env_vars

    resources = args.get("resources")
    # KubernetesPodOperator version 4.2.0 renamed `resources` to
    # `container_resources` (https://github.com/apache/airflow/pull/24673) / (https://github.com/apache/airflow/commit/45f4290712f5f779e57034f81dbaab5d77d5de85)
    # This was done because `KubernetesPodOperator` didn't play nice with dynamic task mapping and they had to
    # deprecate the `resources` argument. Hence, the below code path checks for the version of `KubernetesPodOperator`
    # and then sets the argument. If the version < 4.2.0 then we set the argument as `resources`.
    # If it is > 4.2.0 then we set the argument as `container_resources`
    # The `resources` argument of `KubernetesPodOperator` is going to be deprecated soon in the future.
    # So we will only use it for `KubernetesPodOperator` version < 4.2.0
    # The `resources` argument will also not work for `foreach`s.
    provider_version = get_kubernetes_provider_version()
    k8s_op_ver = create_absolute_version_number(provider_version)
    if k8s_op_ver is None or k8s_op_ver < create_absolute_version_number(
        KUBERNETES_PROVIDER_FOREACH_VERSION
    ):
        # Since the provider version is less than `4.2.0` so we need to use the `resources` argument
        # We need to explicitly parse `resources`/`container_resources` to `k8s.V1ResourceRequirements`,
        # otherwise airflow tries to parse dictionaries to `airflow.providers.cncf.kubernetes.backcompat.pod.Resources`
        # object via `airflow.providers.cncf.kubernetes.backcompat.backward_compat_converts.convert_resources` function.
        # This fails many times since the dictionary structure it expects is not the same as
        # `client.V1ResourceRequirements`.
        args["resources"] = client.V1ResourceRequirements(
            requests=resources["requests"],
            limits=None if "limits" not in resources else resources["limits"],
        )
    else:  # since the provider version is greater than `4.2.0` so should use the `container_resources` argument
        args["container_resources"] = client.V1ResourceRequirements(
            requests=resources["requests"],
            limits=None if "limits" not in resources else resources["limits"],
        )
        del args["resources"]

    if operator_args.get("execution_timeout"):
        args["execution_timeout"] = timedelta(
            **operator_args.get(
                "execution_timeout",
            )
        )
    if operator_args.get("retry_delay"):
        args["retry_delay"] = timedelta(**operator_args.get("retry_delay"))
    return args


def _parse_sensor_args(name, kwargs):
    if name == SensorNames.EXTERNAL_TASK_SENSOR:
        if "execution_delta" in kwargs:
            if type(kwargs["execution_delta"]) == dict:
                kwargs["execution_delta"] = timedelta(**kwargs["execution_delta"])
            else:
                del kwargs["execution_delta"]
    return kwargs


def _get_sensor(name):
    # from airflow import XComArg
    # XComArg()
    if name == SensorNames.EXTERNAL_TASK_SENSOR:
        # ExternalTaskSensors uses an execution_date of a dag to
        # determine the appropriate DAG.
        # This is set to the exact date the current dag gets executed on.
        # For example if "DagA" (Upstream DAG) got scheduled at
        # 12 Jan 4:00 PM PDT then "DagB"(current DAG)'s task sensor will try to
        # look for a "DagA" that got executed at 12 Jan 4:00 PM PDT **exactly**.
        # They also support a `execution_timeout` argument to
        from airflow.sensors.external_task_sensor import ExternalTaskSensor

        return ExternalTaskSensor
    elif name == SensorNames.S3_SENSOR:
        try:
            from airflow.providers.amazon.aws.sensors.s3 import S3KeySensor
        except ImportError:
            raise AirflowSensorNotFound(
                "This DAG requires a `S3KeySensor`. "
                "Install the Airflow AWS provider using : "
                "`pip install apache-airflow-providers-amazon`"
            )
        return S3KeySensor


def get_metaflow_kubernetes_operator():
    try:
        from airflow.contrib.operators.kubernetes_pod_operator import (
            KubernetesPodOperator,
        )
    except ImportError:
        try:
            from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import (
                KubernetesPodOperator,
            )
        except ImportError as e:
            raise KubernetesProviderNotFound(
                "This DAG utilizes `KubernetesPodOperator`. "
                "Install the Airflow Kubernetes provider using "
                "`%s -m pip install apache-airflow-providers-cncf-kubernetes`"
                % sys.executable
            )

    class MetaflowKubernetesOperator(KubernetesPodOperator):
        """
        ## Why Inherit the `KubernetesPodOperator` class ?

        Two key reasons :

        1. So that we can override the `execute` method.
        The only change we introduce to the method is to explicitly modify xcom relating to `return_values`.
        We do this so that the `XComArg` object can work with `expand` function.

        2. So that we can introduce a keyword argument named `mapper_arr`.
        This keyword argument can help as a dummy argument for the `KubernetesPodOperator.partial().expand` method. Any Airflow Operator can be dynamically mapped to runtime artifacts using `Operator.partial(**kwargs).extend(**mapper_kwargs)` post the introduction of [Dynamic Task Mapping](https://airflow.apache.org/docs/apache-airflow/stable/concepts/dynamic-task-mapping.html).
        The `expand` function takes keyword arguments taken by the operator.

        ## Why override the `execute` method  ?

        When we dynamically map vanilla Airflow operators with artifacts generated at runtime, we need to pass that information via `XComArg` to a operator's keyword argument in the `expand` [function](https://airflow.apache.org/docs/apache-airflow/stable/concepts/dynamic-task-mapping.html#mapping-over-result-of-classic-operators).
        The `XComArg` object retrieves XCom values for a particular task based on a `key`, the default key being `return_values`.
        Oddly dynamic task mapping [doesn't support XCom values from any other key except](https://github.com/apache/airflow/blob/8a34d25049a060a035d4db4a49cd4a0d0b07fb0b/airflow/models/mappedoperator.py#L150) `return_values`
        The values of XCom passed by the `KubernetesPodOperator` are mapped to the `return_values` XCom key.

        The biggest problem this creates is that the values of the Foreach cardinality are stored inside the dictionary of `return_values` and cannot be accessed trivially like : `XComArg(task)['foreach_key']` since they are resolved during runtime.
        This puts us in a bind since the only xcom we can retrieve is the full dictionary and we cannot pass that as the iterable for the mapper tasks.
        Hence, we inherit the `execute` method and push custom xcom keys (needed by downstream tasks such as metaflow taskids) and modify `return_values` captured from the container whenever a foreach related xcom is passed.
        When we encounter a foreach xcom we resolve the cardinality which is passed to an actual list and return that as `return_values`.
        This is later useful in the `Workflow.compile` where the operator's `expand` method is called and we are able to retrieve the xcom value.
        """

        template_fields = KubernetesPodOperator.template_fields + (
            "metaflow_pathspec",
            "metaflow_run_id",
            "metaflow_task_id",
            "metaflow_attempt",
            "metaflow_step_name",
            "metaflow_flow_name",
        )

        def __init__(
            self,
            *args,
            mapper_arr=None,
            flow_name=None,
            flow_contains_foreach=False,
            **kwargs
        ) -> None:
            super().__init__(*args, **kwargs)
            self.mapper_arr = mapper_arr
            self._flow_name = flow_name
            self._flow_contains_foreach = flow_contains_foreach
            self.metaflow_pathspec = AIRFLOW_MACROS.pathspec(
                self._flow_name, is_foreach=self._flow_contains_foreach
            )
            self.metaflow_run_id = AIRFLOW_MACROS.RUN_ID
            self.metaflow_task_id = AIRFLOW_MACROS.create_task_id(
                self._flow_contains_foreach
            )
            self.metaflow_attempt = AIRFLOW_MACROS.ATTEMPT
            self.metaflow_step_name = AIRFLOW_MACROS.STEPNAME
            self.metaflow_flow_name = self._flow_name

        def execute(self, context):
            result = super().execute(context)
            if result is None:
                return
            ti = context["ti"]
            if TASK_ID_XCOM_KEY in result:
                ti.xcom_push(
                    key=TASK_ID_XCOM_KEY,
                    value=result[TASK_ID_XCOM_KEY],
                )
            if FOREACH_CARDINALITY_XCOM_KEY in result:
                return list(range(result[FOREACH_CARDINALITY_XCOM_KEY]))

    return MetaflowKubernetesOperator


class AirflowTask(object):
    def __init__(
        self,
        name,
        operator_type="kubernetes",
        flow_name=None,
        is_mapper_node=False,
        flow_contains_foreach=False,
    ):
        self.name = name
        self._is_mapper_node = is_mapper_node
        self._operator_args = None
        self._operator_type = operator_type
        self._flow_name = flow_name
        self._flow_contains_foreach = flow_contains_foreach

    @property
    def is_mapper_node(self):
        return self._is_mapper_node

    def set_operator_args(self, **kwargs):
        self._operator_args = kwargs
        return self

    def _make_sensor(self):
        TaskSensor = _get_sensor(self._operator_type)
        return TaskSensor(
            task_id=self.name,
            **_parse_sensor_args(self._operator_type, self._operator_args)
        )

    def to_dict(self):
        return {
            "name": self.name,
            "is_mapper_node": self._is_mapper_node,
            "operator_type": self._operator_type,
            "operator_args": self._operator_args,
        }

    @classmethod
    def from_dict(cls, task_dict, flow_name=None, flow_contains_foreach=False):
        op_args = {} if "operator_args" not in task_dict else task_dict["operator_args"]
        is_mapper_node = (
            False if "is_mapper_node" not in task_dict else task_dict["is_mapper_node"]
        )
        return cls(
            task_dict["name"],
            is_mapper_node=is_mapper_node,
            operator_type=(
                task_dict["operator_type"]
                if "operator_type" in task_dict
                else "kubernetes"
            ),
            flow_name=flow_name,
            flow_contains_foreach=flow_contains_foreach,
        ).set_operator_args(**op_args)

    def _kubernetes_task(self):
        MetaflowKubernetesOperator = get_metaflow_kubernetes_operator()
        k8s_args = _kubernetes_pod_operator_args(self._operator_args)
        return MetaflowKubernetesOperator(
            flow_name=self._flow_name,
            flow_contains_foreach=self._flow_contains_foreach,
            **k8s_args
        )

    def _kubernetes_mapper_task(self):
        MetaflowKubernetesOperator = get_metaflow_kubernetes_operator()
        k8s_args = _kubernetes_pod_operator_args(self._operator_args)
        return MetaflowKubernetesOperator.partial(
            flow_name=self._flow_name,
            flow_contains_foreach=self._flow_contains_foreach,
            **k8s_args
        )

    def to_task(self):
        if self._operator_type == "kubernetes":
            if not self.is_mapper_node:
                return self._kubernetes_task()
            else:
                return self._kubernetes_mapper_task()
        elif self._operator_type in SensorNames.get_supported_sensors():
            return self._make_sensor()


class Workflow(object):
    def __init__(self, file_path=None, graph_structure=None, metadata=None, **kwargs):
        self._dag_instantiation_params = AirflowDAGArgs(**kwargs)
        self._file_path = file_path
        self._metadata = metadata
        tree = lambda: defaultdict(tree)
        self.states = tree()
        self.metaflow_params = None
        self.graph_structure = graph_structure

    def set_parameters(self, params):
        self.metaflow_params = params

    def add_state(self, state):
        self.states[state.name] = state

    def to_dict(self):
        return dict(
            metadata=self._metadata,
            graph_structure=self.graph_structure,
            states={s: v.to_dict() for s, v in self.states.items()},
            dag_instantiation_params=self._dag_instantiation_params.serialize(),
            file_path=self._file_path,
            metaflow_params=self.metaflow_params,
        )

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data_dict):
        re_cls = cls(
            file_path=data_dict["file_path"],
            graph_structure=data_dict["graph_structure"],
            metadata=data_dict["metadata"],
        )
        re_cls._dag_instantiation_params = AirflowDAGArgs.deserialize(
            data_dict["dag_instantiation_params"]
        )

        for sd in data_dict["states"].values():
            re_cls.add_state(
                AirflowTask.from_dict(sd, flow_name=data_dict["metadata"]["flow_name"])
            )
        re_cls.set_parameters(data_dict["metaflow_params"])
        return re_cls

    @classmethod
    def from_json(cls, json_string):
        data = json.loads(json_string)
        return cls.from_dict(data)

    def _construct_params(self):
        from airflow.models.param import Param

        if self.metaflow_params is None:
            return {}
        param_dict = {}
        for p in self.metaflow_params:
            name = p["name"]
            del p["name"]
            param_dict[name] = Param(**p)
        return param_dict

    def compile(self):
        from airflow import DAG

        # Airflow 2.0.0 cannot import this, so we have to do it this way.
        # `XComArg` is needed for dynamic task mapping and if the airflow installation is of the right
        # version (+2.3.0) then the class will be importable.
        XComArg = get_xcom_arg_class()

        _validate_minimum_airflow_version()

        if self._metadata["contains_foreach"]:
            _validate_dynamic_mapping_compatibility()
            # We need to verify if KubernetesPodOperator is of version > 4.2.0 to support foreachs / dynamic task mapping.
            # If the dag uses dynamic Task mapping then we throw an error since the `resources` argument in the `KubernetesPodOperator`
            # doesn't work for dynamic task mapping for `KubernetesPodOperator` version < 4.2.0.
            # For more context check this issue :  https://github.com/apache/airflow/issues/24669
            _check_foreach_compatible_kubernetes_provider()

        params_dict = self._construct_params()
        # DAG Params can be seen here :
        # https://airflow.apache.org/docs/apache-airflow/2.0.0/_api/airflow/models/dag/index.html#airflow.models.dag.DAG
        # Airflow 2.0.0 Allows setting Params.
        dag = DAG(params=params_dict, **self._dag_instantiation_params.arguments)
        dag.fileloc = self._file_path if self._file_path is not None else dag.fileloc

        def add_node(node, parents, dag):
            """
            A recursive function to traverse the specialized
            graph_structure datastructure.
            """
            if type(node) == str:
                task = self.states[node].to_task()
                if parents:
                    for parent in parents:
                        # Handle foreach nodes.
                        if self.states[node].is_mapper_node:
                            task = task.expand(mapper_arr=XComArg(parent))
                        parent >> task
                return [task]  # Return Parent

            # this means a split from parent
            if type(node) == list:
                # this means branching since everything within the list is a list
                if all(isinstance(n, list) for n in node):
                    curr_parents = parents
                    parent_list = []
                    for node_list in node:
                        last_parent = add_node(node_list, curr_parents, dag)
                        parent_list.extend(last_parent)
                    return parent_list
                else:
                    # this means no branching and everything within the list is not a list and can be actual nodes.
                    curr_parents = parents
                    for node_x in node:
                        curr_parents = add_node(node_x, curr_parents, dag)
                    return curr_parents

        with dag:
            parent = None
            for node in self.graph_structure:
                parent = add_node(node, parent, dag)

        return dag



import os
import sys
import types

from metaflow.exception import MetaflowException
from metaflow.metaflow_config_funcs import from_conf, get_validate_choice_fn

# Disable multithreading security on MacOS
if sys.platform == "darwin":
    os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"

## NOTE: Just like Click's auto_envar_prefix `METAFLOW` (see in cli.py), all environment
## variables here are also named METAFLOW_XXX. So, for example, in the statement:
## `DEFAULT_DATASTORE = from_conf("DEFAULT_DATASTORE", "local")`, to override the default
## value, either set `METAFLOW_DEFAULT_DATASTORE` in your configuration file or set
## an environment variable called `METAFLOW_DEFAULT_DATASTORE`

##
# Constants (NOTE: these need to live before any from_conf)
##

# Path to the local directory to store artifacts for 'local' datastore.
DATASTORE_LOCAL_DIR = ".metaflow"

# Local configuration file (in .metaflow) containing overrides per-project
LOCAL_CONFIG_FILE = "config.json"

###
# Default configuration
###

DEFAULT_DATASTORE = from_conf("DEFAULT_DATASTORE", "local")
DEFAULT_ENVIRONMENT = from_conf("DEFAULT_ENVIRONMENT", "local")
DEFAULT_EVENT_LOGGER = from_conf("DEFAULT_EVENT_LOGGER", "nullSidecarLogger")
DEFAULT_METADATA = from_conf("DEFAULT_METADATA", "local")
DEFAULT_MONITOR = from_conf("DEFAULT_MONITOR", "nullSidecarMonitor")
DEFAULT_PACKAGE_SUFFIXES = from_conf("DEFAULT_PACKAGE_SUFFIXES", ".py,.R,.RDS")
DEFAULT_AWS_CLIENT_PROVIDER = from_conf("DEFAULT_AWS_CLIENT_PROVIDER", "boto3")
DEFAULT_AZURE_CLIENT_PROVIDER = from_conf(
    "DEFAULT_AZURE_CLIENT_PROVIDER", "azure-default"
)
DEFAULT_GCP_CLIENT_PROVIDER = from_conf("DEFAULT_GCP_CLIENT_PROVIDER", "gcp-default")
DEFAULT_SECRETS_BACKEND_TYPE = from_conf("DEFAULT_SECRETS_BACKEND_TYPE")
DEFAULT_SECRETS_ROLE = from_conf("DEFAULT_SECRETS_ROLE")

DEFAULT_FROM_DEPLOYMENT_IMPL = from_conf(
    "DEFAULT_FROM_DEPLOYMENT_IMPL", "argo-workflows"
)

###
# User configuration
###
USER = from_conf("USER")


###
# Datastore configuration
###
DATASTORE_SYSROOT_LOCAL = from_conf("DATASTORE_SYSROOT_LOCAL")
# S3 bucket and prefix to store artifacts for 's3' datastore.
DATASTORE_SYSROOT_S3 = from_conf("DATASTORE_SYSROOT_S3")
# Azure Blob Storage container and blob prefix
DATASTORE_SYSROOT_AZURE = from_conf("DATASTORE_SYSROOT_AZURE")
DATASTORE_SYSROOT_GS = from_conf("DATASTORE_SYSROOT_GS")
# GS bucket and prefix to store artifacts for 'gs' datastore


###
# Datastore local cache
###

# Path to the client cache
CLIENT_CACHE_PATH = from_conf("CLIENT_CACHE_PATH", "/tmp/metaflow_client")
# Maximum size (in bytes) of the cache
CLIENT_CACHE_MAX_SIZE = from_conf("CLIENT_CACHE_MAX_SIZE", 10000)
# Maximum number of cached Flow and TaskDatastores in the cache
CLIENT_CACHE_MAX_FLOWDATASTORE_COUNT = from_conf(
    "CLIENT_CACHE_MAX_FLOWDATASTORE_COUNT", 50
)
CLIENT_CACHE_MAX_TASKDATASTORE_COUNT = from_conf(
    "CLIENT_CACHE_MAX_TASKDATASTORE_COUNT", CLIENT_CACHE_MAX_FLOWDATASTORE_COUNT * 100
)


###
# Datatools (S3) configuration
###
S3_ENDPOINT_URL = from_conf("S3_ENDPOINT_URL")
S3_VERIFY_CERTIFICATE = from_conf("S3_VERIFY_CERTIFICATE")

# Set ServerSideEncryption for S3 uploads
S3_SERVER_SIDE_ENCRYPTION = from_conf("S3_SERVER_SIDE_ENCRYPTION")

# S3 retry configuration
# This is useful if you want to "fail fast" on S3 operations; use with caution
# though as this may increase failures. Note that this is the number of *retries*
# so setting it to 0 means each operation will be tried once.
S3_RETRY_COUNT = from_conf("S3_RETRY_COUNT", 7)

# Number of concurrent S3 processes for parallel operations.
S3_WORKER_COUNT = from_conf("S3_WORKER_COUNT", 64)

# Number of retries on *transient* failures (such as SlowDown errors). Note
# that if after S3_TRANSIENT_RETRY_COUNT times, all operations haven't been done,
# it will try up to S3_RETRY_COUNT again so the total number of tries can be up to
# (S3_RETRY_COUNT + 1) * (S3_TRANSIENT_RETRY_COUNT + 1)
# You typically want this number fairly high as transient retires are "cheap" (only
# operations that have not succeeded retry as opposed to all operations for the
# top-level retries)
S3_TRANSIENT_RETRY_COUNT = from_conf("S3_TRANSIENT_RETRY_COUNT", 20)

# S3 retry configuration used in the aws client
# Use the adaptive retry strategy by default
S3_CLIENT_RETRY_CONFIG = from_conf(
    "S3_CLIENT_RETRY_CONFIG", {"max_attempts": 10, "mode": "adaptive"}
)

# Threshold to start printing warnings for an AWS retry
RETRY_WARNING_THRESHOLD = 3

# S3 datatools root location
DATATOOLS_SUFFIX = from_conf("DATATOOLS_SUFFIX", "data")
DATATOOLS_S3ROOT = from_conf(
    "DATATOOLS_S3ROOT",
    (
        os.path.join(DATASTORE_SYSROOT_S3, DATATOOLS_SUFFIX)
        if DATASTORE_SYSROOT_S3
        else None
    ),
)

TEMPDIR = from_conf("TEMPDIR", ".")

DATATOOLS_CLIENT_PARAMS = from_conf("DATATOOLS_CLIENT_PARAMS", {})
if S3_ENDPOINT_URL:
    DATATOOLS_CLIENT_PARAMS["endpoint_url"] = S3_ENDPOINT_URL
if S3_VERIFY_CERTIFICATE:
    DATATOOLS_CLIENT_PARAMS["verify"] = S3_VERIFY_CERTIFICATE

DATATOOLS_SESSION_VARS = from_conf("DATATOOLS_SESSION_VARS", {})

# Azure datatools root location
# Note: we do not expose an actual datatools library for Azure (like we do for S3)
# Similar to DATATOOLS_LOCALROOT, this is used ONLY by the IncludeFile's internal implementation.
DATATOOLS_AZUREROOT = from_conf(
    "DATATOOLS_AZUREROOT",
    (
        os.path.join(DATASTORE_SYSROOT_AZURE, DATATOOLS_SUFFIX)
        if DATASTORE_SYSROOT_AZURE
        else None
    ),
)
# GS datatools root location
# Note: we do not expose an actual datatools library for GS (like we do for S3)
# Similar to DATATOOLS_LOCALROOT, this is used ONLY by the IncludeFile's internal implementation.
DATATOOLS_GSROOT = from_conf(
    "DATATOOLS_GSROOT",
    (
        os.path.join(DATASTORE_SYSROOT_GS, DATATOOLS_SUFFIX)
        if DATASTORE_SYSROOT_GS
        else None
    ),
)
# Local datatools root location
DATATOOLS_LOCALROOT = from_conf(
    "DATATOOLS_LOCALROOT",
    (
        os.path.join(DATASTORE_SYSROOT_LOCAL, DATATOOLS_SUFFIX)
        if DATASTORE_SYSROOT_LOCAL
        else None
    ),
)

# Secrets Backend - AWS Secrets Manager configuration
AWS_SECRETS_MANAGER_DEFAULT_REGION = from_conf("AWS_SECRETS_MANAGER_DEFAULT_REGION")

# Secrets Backend - GCP Secrets name prefix. With this, users don't have
# to specify the full secret name in the @secret decorator.
#
# Note that it makes a difference whether the prefix ends with a slash or not
# E.g. if secret name passed to @secret decorator is mysecret:
# - "projects/1234567890/secrets/" -> "projects/1234567890/secrets/mysecret"
# - "projects/1234567890/secrets/foo-" -> "projects/1234567890/secrets/foo-mysecret"
GCP_SECRET_MANAGER_PREFIX = from_conf("GCP_SECRET_MANAGER_PREFIX")

# Secrets Backend - Azure Key Vault prefix. With this, users don't have to
# specify the full https:// vault url in the @secret decorator.
#
# It does not make a difference if the prefix ends in a / or not. We will handle either
# case correctly.
AZURE_KEY_VAULT_PREFIX = from_conf("AZURE_KEY_VAULT_PREFIX")

# The root directory to save artifact pulls in, when using S3 or Azure
ARTIFACT_LOCALROOT = from_conf("ARTIFACT_LOCALROOT", os.getcwd())

# Cards related config variables
CARD_SUFFIX = "mf.cards"
CARD_LOCALROOT = from_conf("CARD_LOCALROOT")
CARD_S3ROOT = from_conf(
    "CARD_S3ROOT",
    os.path.join(DATASTORE_SYSROOT_S3, CARD_SUFFIX) if DATASTORE_SYSROOT_S3 else None,
)
CARD_AZUREROOT = from_conf(
    "CARD_AZUREROOT",
    (
        os.path.join(DATASTORE_SYSROOT_AZURE, CARD_SUFFIX)
        if DATASTORE_SYSROOT_AZURE
        else None
    ),
)
CARD_GSROOT = from_conf(
    "CARD_GSROOT",
    os.path.join(DATASTORE_SYSROOT_GS, CARD_SUFFIX) if DATASTORE_SYSROOT_GS else None,
)
CARD_NO_WARNING = from_conf("CARD_NO_WARNING", False)

RUNTIME_CARD_RENDER_INTERVAL = from_conf("RUNTIME_CARD_RENDER_INTERVAL", 60)

# Azure storage account URL
AZURE_STORAGE_BLOB_SERVICE_ENDPOINT = from_conf("AZURE_STORAGE_BLOB_SERVICE_ENDPOINT")

# Azure storage can use process-based parallelism instead of threads.
# Processes perform better for high throughput workloads (e.g. many huge artifacts)
AZURE_STORAGE_WORKLOAD_TYPE = from_conf(
    "AZURE_STORAGE_WORKLOAD_TYPE",
    default="general",
    validate_fn=get_validate_choice_fn(["general", "high_throughput"]),
)

# GS storage can use process-based parallelism instead of threads.
# Processes perform better for high throughput workloads (e.g. many huge artifacts)
GS_STORAGE_WORKLOAD_TYPE = from_conf(
    "GS_STORAGE_WORKLOAD_TYPE",
    "general",
    validate_fn=get_validate_choice_fn(["general", "high_throughput"]),
)

###
# Metadata configuration
###
SERVICE_URL = from_conf("SERVICE_URL")
SERVICE_RETRY_COUNT = from_conf("SERVICE_RETRY_COUNT", 5)
SERVICE_AUTH_KEY = from_conf("SERVICE_AUTH_KEY")
SERVICE_HEADERS = from_conf("SERVICE_HEADERS", {})
if SERVICE_AUTH_KEY is not None:
    SERVICE_HEADERS["x-api-key"] = SERVICE_AUTH_KEY
# Checks version compatibility with Metadata service
SERVICE_VERSION_CHECK = from_conf("SERVICE_VERSION_CHECK", True)

# Default container image
DEFAULT_CONTAINER_IMAGE = from_conf("DEFAULT_CONTAINER_IMAGE")
# Default container registry
DEFAULT_CONTAINER_REGISTRY = from_conf("DEFAULT_CONTAINER_REGISTRY")
# Controls whether to include foreach stack information in metadata.
INCLUDE_FOREACH_STACK = from_conf("INCLUDE_FOREACH_STACK", True)
# Maximum length of the foreach value string to be stored in each ForeachFrame.
MAXIMUM_FOREACH_VALUE_CHARS = from_conf("MAXIMUM_FOREACH_VALUE_CHARS", 30)
# The default runtime limit (In seconds) of jobs launched by any compute provider. Default of 5 days.
DEFAULT_RUNTIME_LIMIT = from_conf("DEFAULT_RUNTIME_LIMIT", 5 * 24 * 60 * 60)

###
# Organization customizations
###
UI_URL = from_conf("UI_URL")
PAGERDUTY_TEMPLATE_URL = from_conf("PAGERDUTY_TEMPLATE_URL")

###
# Capture error logs from argo
###
ARGO_WORKFLOWS_CAPTURE_ERROR_SCRIPT = from_conf("ARGO_WORKFLOWS_CAPTURE_ERROR_SCRIPT")

# Contact information displayed when running the `metaflow` command.
# Value should be a dictionary where:
#  - key is a string describing contact method
#  - value is a string describing contact itself (email, web address, etc.)
# The default value shows an example of this
CONTACT_INFO = from_conf(
    "CONTACT_INFO",
    {
        "Read the documentation": "http://docs.metaflow.org",
        "Chat with us": "http://chat.metaflow.org",
        "Get help by email": "help@metaflow.org",
    },
)


###
# Decorators
###
# Format is a space separated string of decospecs (what is passed
# using --with)
DEFAULT_DECOSPECS = from_conf("DEFAULT_DECOSPECS", "")

###
# AWS Batch configuration
###
# IAM role for AWS Batch container with Amazon S3 access
# (and AWS DynamoDb access for AWS StepFunctions, if enabled)
ECS_S3_ACCESS_IAM_ROLE = from_conf("ECS_S3_ACCESS_IAM_ROLE")
# IAM role for AWS Batch container for AWS Fargate
ECS_FARGATE_EXECUTION_ROLE = from_conf("ECS_FARGATE_EXECUTION_ROLE")
# Job queue for AWS Batch
BATCH_JOB_QUEUE = from_conf("BATCH_JOB_QUEUE")
# Default container image for AWS Batch
BATCH_CONTAINER_IMAGE = from_conf("BATCH_CONTAINER_IMAGE", DEFAULT_CONTAINER_IMAGE)
# Default container registry for AWS Batch
BATCH_CONTAINER_REGISTRY = from_conf(
    "BATCH_CONTAINER_REGISTRY", DEFAULT_CONTAINER_REGISTRY
)
# Metadata service URL for AWS Batch
SERVICE_INTERNAL_URL = from_conf("SERVICE_INTERNAL_URL", SERVICE_URL)

# Assign resource tags to AWS Batch jobs. Set to False by default since
# it requires `Batch:TagResource` permissions which may not be available
# in all Metaflow deployments. Hopefully, some day we can flip the
# default to True.
BATCH_EMIT_TAGS = from_conf("BATCH_EMIT_TAGS", False)
# Default tags to add to AWS Batch jobs. These are in addition to the defaults set when BATCH_EMIT_TAGS is true.
BATCH_DEFAULT_TAGS = from_conf("BATCH_DEFAULT_TAGS", {})

###
# AWS Step Functions configuration
###
# IAM role for AWS Step Functions with AWS Batch and AWS DynamoDb access
# https://docs.aws.amazon.com/step-functions/latest/dg/batch-iam.html
SFN_IAM_ROLE = from_conf("SFN_IAM_ROLE")
# AWS DynamoDb Table name (with partition key - `pathspec` of type string)
SFN_DYNAMO_DB_TABLE = from_conf("SFN_DYNAMO_DB_TABLE")
# IAM role for AWS Events with AWS Step Functions access
# https://docs.aws.amazon.com/eventbridge/latest/userguide/auth-and-access-control-eventbridge.html
EVENTS_SFN_ACCESS_IAM_ROLE = from_conf("EVENTS_SFN_ACCESS_IAM_ROLE")
# Prefix for AWS Step Functions state machines. Set to stack name for Metaflow
# sandbox.
SFN_STATE_MACHINE_PREFIX = from_conf("SFN_STATE_MACHINE_PREFIX")
# Optional AWS CloudWatch Log Group ARN for emitting AWS Step Functions state
# machine execution logs. This needs to be available when using the
# `step-functions create --log-execution-history` command.
SFN_EXECUTION_LOG_GROUP_ARN = from_conf("SFN_EXECUTION_LOG_GROUP_ARN")
# Amazon S3 path for storing the results of AWS Step Functions Distributed Map
SFN_S3_DISTRIBUTED_MAP_OUTPUT_PATH = from_conf(
    "SFN_S3_DISTRIBUTED_MAP_OUTPUT_PATH",
    (
        os.path.join(DATASTORE_SYSROOT_S3, "sfn_distributed_map_output")
        if DATASTORE_SYSROOT_S3
        else None
    ),
)
# Toggle for step command being part of the Step Function payload, or if it should be offloaded to S3
SFN_COMPRESS_STATE_MACHINE = from_conf("SFN_COMPRESS_STATE_MACHINE", False)
###
# Kubernetes configuration
###
# Kubernetes namespace to use for all objects created by Metaflow
KUBERNETES_NAMESPACE = from_conf("KUBERNETES_NAMESPACE", "default")
# Default service account to use by K8S jobs created by Metaflow
KUBERNETES_SERVICE_ACCOUNT = from_conf("KUBERNETES_SERVICE_ACCOUNT")
# Default node selectors to use by K8S jobs created by Metaflow - foo=bar,baz=bab
KUBERNETES_NODE_SELECTOR = from_conf("KUBERNETES_NODE_SELECTOR", "")
KUBERNETES_TOLERATIONS = from_conf("KUBERNETES_TOLERATIONS", "")
KUBERNETES_PERSISTENT_VOLUME_CLAIMS = from_conf(
    "KUBERNETES_PERSISTENT_VOLUME_CLAIMS", ""
)
KUBERNETES_SECRETS = from_conf("KUBERNETES_SECRETS", "")
# Default labels for kubernetes pods
KUBERNETES_LABELS = from_conf("KUBERNETES_LABELS", "")
# Default annotations for kubernetes pods
KUBERNETES_ANNOTATIONS = from_conf("KUBERNETES_ANNOTATIONS", "")
# Default GPU vendor to use by K8S jobs created by Metaflow (supports nvidia, amd)
KUBERNETES_GPU_VENDOR = from_conf("KUBERNETES_GPU_VENDOR", "nvidia")
# Default container image for K8S
KUBERNETES_CONTAINER_IMAGE = from_conf(
    "KUBERNETES_CONTAINER_IMAGE", DEFAULT_CONTAINER_IMAGE
)
# Image pull policy for container images
KUBERNETES_IMAGE_PULL_POLICY = from_conf("KUBERNETES_IMAGE_PULL_POLICY", None)
# Image pull secrets for container images
KUBERNETES_IMAGE_PULL_SECRETS = from_conf("KUBERNETES_IMAGE_PULL_SECRETS", "")
# Default container registry for K8S
KUBERNETES_CONTAINER_REGISTRY = from_conf(
    "KUBERNETES_CONTAINER_REGISTRY", DEFAULT_CONTAINER_REGISTRY
)
# Toggle for trying to fetch EC2 instance metadata
KUBERNETES_FETCH_EC2_METADATA = from_conf("KUBERNETES_FETCH_EC2_METADATA", False)
# Shared memory in MB to use for this step
KUBERNETES_SHARED_MEMORY = from_conf("KUBERNETES_SHARED_MEMORY", None)
# Default port number to open on the pods
KUBERNETES_PORT = from_conf("KUBERNETES_PORT", None)
# Default kubernetes resource requests for CPU, memory and disk
KUBERNETES_CPU = from_conf("KUBERNETES_CPU", None)
KUBERNETES_MEMORY = from_conf("KUBERNETES_MEMORY", None)
KUBERNETES_DISK = from_conf("KUBERNETES_DISK", None)
# Default kubernetes QoS class
KUBERNETES_QOS = from_conf("KUBERNETES_QOS", "burstable")

# Architecture of kubernetes nodes - used for @conda/@pypi in metaflow-dev
KUBERNETES_CONDA_ARCH = from_conf("KUBERNETES_CONDA_ARCH")
ARGO_WORKFLOWS_KUBERNETES_SECRETS = from_conf("ARGO_WORKFLOWS_KUBERNETES_SECRETS", "")
ARGO_WORKFLOWS_ENV_VARS_TO_SKIP = from_conf("ARGO_WORKFLOWS_ENV_VARS_TO_SKIP", "")

KUBERNETES_JOBSET_GROUP = from_conf("KUBERNETES_JOBSET_GROUP", "jobset.x-k8s.io")
KUBERNETES_JOBSET_VERSION = from_conf("KUBERNETES_JOBSET_VERSION", "v1alpha2")

##
# Argo Events Configuration
##
ARGO_EVENTS_SERVICE_ACCOUNT = from_conf("ARGO_EVENTS_SERVICE_ACCOUNT")
ARGO_EVENTS_EVENT_BUS = from_conf("ARGO_EVENTS_EVENT_BUS", "default")
ARGO_EVENTS_EVENT_SOURCE = from_conf("ARGO_EVENTS_EVENT_SOURCE")
ARGO_EVENTS_EVENT = from_conf("ARGO_EVENTS_EVENT")
ARGO_EVENTS_WEBHOOK_URL = from_conf("ARGO_EVENTS_WEBHOOK_URL")
ARGO_EVENTS_INTERNAL_WEBHOOK_URL = from_conf(
    "ARGO_EVENTS_INTERNAL_WEBHOOK_URL", ARGO_EVENTS_WEBHOOK_URL
)
ARGO_EVENTS_WEBHOOK_AUTH = from_conf("ARGO_EVENTS_WEBHOOK_AUTH", "none")
ARGO_EVENTS_SENSOR_NAMESPACE = from_conf(
    "ARGO_EVENTS_SENSOR_NAMESPACE", KUBERNETES_NAMESPACE
)

ARGO_WORKFLOWS_UI_URL = from_conf("ARGO_WORKFLOWS_UI_URL")

##
# Airflow Configuration
##
# This configuration sets `startup_timeout_seconds` in airflow's KubernetesPodOperator.
AIRFLOW_KUBERNETES_STARTUP_TIMEOUT_SECONDS = from_conf(
    "AIRFLOW_KUBERNETES_STARTUP_TIMEOUT_SECONDS", 60 * 60
)
# This configuration sets `kubernetes_conn_id` in airflow's KubernetesPodOperator.
AIRFLOW_KUBERNETES_CONN_ID = from_conf("AIRFLOW_KUBERNETES_CONN_ID")
AIRFLOW_KUBERNETES_KUBECONFIG_FILE = from_conf("AIRFLOW_KUBERNETES_KUBECONFIG_FILE")
AIRFLOW_KUBERNETES_KUBECONFIG_CONTEXT = from_conf(
    "AIRFLOW_KUBERNETES_KUBECONFIG_CONTEXT"
)


###
# Conda configuration
###
# Conda package root location on S3
CONDA_PACKAGE_S3ROOT = from_conf("CONDA_PACKAGE_S3ROOT")
# Conda package root location on Azure
CONDA_PACKAGE_AZUREROOT = from_conf("CONDA_PACKAGE_AZUREROOT")
# Conda package root location on GS
CONDA_PACKAGE_GSROOT = from_conf("CONDA_PACKAGE_GSROOT")

# Use an alternate dependency resolver for conda packages instead of conda
# Mamba promises faster package dependency resolution times, which
# should result in an appreciable speedup in flow environment initialization.
CONDA_DEPENDENCY_RESOLVER = from_conf("CONDA_DEPENDENCY_RESOLVER", "conda")

# Default to not using fast init binary.
CONDA_USE_FAST_INIT = from_conf("CONDA_USE_FAST_INIT", False)

###
# Escape hatch configuration
###
# Print out warning if escape hatch is not used for the target packages
ESCAPE_HATCH_WARNING = from_conf("ESCAPE_HATCH_WARNING", True)

###
# Features
###
FEAT_ALWAYS_UPLOAD_CODE_PACKAGE = from_conf("FEAT_ALWAYS_UPLOAD_CODE_PACKAGE", False)
###
# Debug configuration
###
DEBUG_OPTIONS = [
    "subcommand",
    "sidecar",
    "s3client",
    "tracing",
    "stubgen",
    "userconf",
    "conda",
    "package",
]

for typ in DEBUG_OPTIONS:
    vars()["DEBUG_%s" % typ.upper()] = from_conf("DEBUG_%s" % typ.upper(), False)

###
# Plugin configuration
###

# Plugin configuration variables exist in plugins/__init__.py.
# Specifically, there is an ENABLED_<category> configuration value to determine
# the set of plugins to enable. The categories are: step_decorator, flow_decorator,
# environment, metadata_provider, datastore, sidecar, logging_sidecar, monitor_sidecar,
# aws_client_provider, and cli. If not set (the default), all plugins are enabled.
# You can restrict which plugins are enabled by listing them explicitly, for example
# ENABLED_STEP_DECORATOR = ["batch", "resources"] will enable only those two step
# decorators and none other.

###
# Command configuration
###

# Command (ie: metaflow <cmd>) configuration variable ENABLED_CMD
# exists in cmd/main_cli.py. It behaves just like any of the other ENABLED_<category>
# configuration variables.

###
# AWS Sandbox configuration
###
# Boolean flag for metaflow AWS sandbox access
AWS_SANDBOX_ENABLED = from_conf("AWS_SANDBOX_ENABLED", False)
# Metaflow AWS sandbox auth endpoint
AWS_SANDBOX_STS_ENDPOINT_URL = SERVICE_URL
# Metaflow AWS sandbox API auth key
AWS_SANDBOX_API_KEY = from_conf("AWS_SANDBOX_API_KEY")
# Internal Metadata URL
AWS_SANDBOX_INTERNAL_SERVICE_URL = from_conf("AWS_SANDBOX_INTERNAL_SERVICE_URL")
# AWS region
AWS_SANDBOX_REGION = from_conf("AWS_SANDBOX_REGION")


# Finalize configuration
if AWS_SANDBOX_ENABLED:
    os.environ["AWS_DEFAULT_REGION"] = AWS_SANDBOX_REGION
    SERVICE_INTERNAL_URL = AWS_SANDBOX_INTERNAL_SERVICE_URL
    SERVICE_HEADERS["x-api-key"] = AWS_SANDBOX_API_KEY
    SFN_STATE_MACHINE_PREFIX = from_conf("AWS_SANDBOX_STACK_NAME")

KUBERNETES_SANDBOX_INIT_SCRIPT = from_conf("KUBERNETES_SANDBOX_INIT_SCRIPT")

OTEL_ENDPOINT = from_conf("OTEL_ENDPOINT")
ZIPKIN_ENDPOINT = from_conf("ZIPKIN_ENDPOINT")
CONSOLE_TRACE_ENABLED = from_conf("CONSOLE_TRACE_ENABLED", False)
# internal env used for preventing the tracing module from loading during Conda bootstrapping.
DISABLE_TRACING = bool(os.environ.get("DISABLE_TRACING", False))
MAX_MEMORY_PER_TASK = from_conf("MAX_MEMORY_PER_TASK")
MAX_CPU_PER_TASK = from_conf("MAX_CPU_PER_TASK")

# MAX_ATTEMPTS is the maximum number of attempts, including the first
# task, retries, and the final fallback task and its retries.
#
# Datastore needs to check all attempt files to find the latest one, so
# increasing this limit has real performance implications for all tasks.
# Decreasing this limit is very unsafe, as it can lead to wrong results
# being read from old tasks.
#
# Note also that DataStoreSet resolves the latest attempt_id using
# lexicographic ordering of attempts. This won't work if MAX_ATTEMPTS > 99.
MAX_ATTEMPTS = 6

# Feature flag (experimental features that are *explicitly* unsupported)

# Process configs even when using the click_api for Runner/Deployer
CLICK_API_PROCESS_CONFIG = from_conf("CLICK_API_PROCESS_CONFIG", False)


# PINNED_CONDA_LIBS are the libraries that metaflow depends on for execution
# and are needed within a conda environment
def get_pinned_conda_libs(python_version, datastore_type):
    pins = {
        "requests": ">=2.21.0",
    }
    if datastore_type == "s3":
        pins["boto3"] = ">=1.14.0"
    elif datastore_type == "azure":
        pins["azure-identity"] = ">=1.10.0"
        pins["azure-storage-blob"] = ">=12.12.0"
        pins["azure-keyvault-secrets"] = ">=4.7.0"
        pins["simple-azure-blob-downloader"] = ">=0.1.0"
    elif datastore_type == "gs":
        pins["google-cloud-storage"] = ">=2.5.0"
        pins["google-auth"] = ">=2.11.0"
        pins["google-cloud-secret-manager"] = ">=2.10.0"
        pins["simple-gcp-object-downloader"] = ">=0.1.0"
    elif datastore_type == "local":
        pass
    else:
        raise MetaflowException(
            msg="conda lib pins for datastore %s are undefined" % (datastore_type,)
        )
    return pins


# Check if there are extensions to Metaflow to load and override everything
try:
    from metaflow.extension_support import get_modules

    _TOGGLE_DECOSPECS = []

    ext_modules = get_modules("config")
    for m in ext_modules:
        # We load into globals whatever we have in extension_module
        # We specifically exclude any modules that may be included (like sys, os, etc)
        for n, o in m.module.__dict__.items():
            if n == "DEBUG_OPTIONS":
                DEBUG_OPTIONS.extend(o)
                for typ in o:
                    vars()["DEBUG_%s" % typ.upper()] = from_conf(
                        "DEBUG_%s" % typ.upper(), False
                    )
            elif n == "get_pinned_conda_libs":

                def _new_get_pinned_conda_libs(
                    python_version, datastore_type, f1=globals()[n], f2=o
                ):
                    d1 = f1(python_version, datastore_type)
                    d2 = f2(python_version, datastore_type)
                    for k, v in d2.items():
                        d1[k] = v if k not in d1 else ",".join([d1[k], v])
                    return d1

                globals()[n] = _new_get_pinned_conda_libs
            elif n == "TOGGLE_DECOSPECS":
                if any([x.startswith("-") for x in o]):
                    raise ValueError("Removing decospecs is not currently supported")
                if any(" " in x for x in o):
                    raise ValueError("Decospecs cannot contain spaces")
                _TOGGLE_DECOSPECS.extend(o)
            elif not n.startswith("__") and not isinstance(o, types.ModuleType):
                globals()[n] = o
    # If DEFAULT_DECOSPECS is set, use that, else extrapolate from extensions
    if not DEFAULT_DECOSPECS:
        DEFAULT_DECOSPECS = " ".join(_TOGGLE_DECOSPECS)

finally:
    # Erase all temporary names to avoid leaking things
    for _n in [
        "m",
        "n",
        "o",
        "typ",
        "ext_modules",
        "get_modules",
        "_new_get_pinned_conda_libs",
        "d1",
        "d2",
        "k",
        "v",
        "f1",
        "f2",
        "_TOGGLE_DECOSPECS",
    ]:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()["_n"]
